/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"

int io_in = 0;

io_err_spec_used (spec)
char *spec;
{
	sprintf (serr, "specifier %s already used", spec);
	fatal (serr);
}

io_err_spec_type (spec, type)
char *spec, *type;
{
	sprintf (serr, "expression for specifier %s must be of type %s", spec, type);
	fatal (serr);
}

int io_spec_is_name (key, name, type, spec, unitspec)
char *key, *name, *type;
struct S_IoSpec *spec;
Node *unitspec;
{
	int ret;

	ret = io_spec_is (key, name, type, spec, unitspec);

	if (ret && spec->Expr->com.type != NAME) {
		sprintf (serr, "specifier %s needs a variable", key);
		fatal (serr);
	}

	return ret;
}

int io_spec_is (key, name, type, spec, unitspec)
char *key, *name, *type;
struct S_IoSpec *spec;
Node *unitspec;
{
	if (strncmp (name, key, strlen (key))) {
		return 0;
	}

	if (unitspec != NN) {
		io_err_spec_used (key);
		return 1;
	}

	if (!strncmp (type, "CHARACTER", 9)) {
		if (!IS_CHARACTER (spec->Expr)) {
			io_err_spec_type (key, type);
			return 1;
		}
	}
	else if (!strncmp (type, "INTEGER", 9)) {
		if (!IS_INTEGER (spec->Expr)) {
			io_err_spec_type (key, type);
			return 1;
		}
	}

	return 1;
}

/********************************* C H E C K ********************************/
Node *check_WriteStmt (root)
Node *root;
{
	return check_IoStmt (root, IO_IS_WRITE);
}

Node *check_ReadStmt (root)
Node *root;
{
	return check_IoStmt (root, IO_IS_READ);
}

Node *check_PrintStmt (root)
Node *root;
{
	return check_IoStmt (root, IO_IS_PRINT);
}

Node *check_IoStmt (root, io_type)
Node *root;
int io_type;
{
	Node *list, *nspec, *temp;
	Unit *unit = unit_malloc ();
	struct S_IoSpec *spec;
	int spec_num;
	char *name;
	T_Entry *entry;

/*	if (flag_cubix)  BOZKUS 
		cubix_io_called = 1; */

	switch (io_type) {

	case IO_IS_WRITE:
		root->WriteStmt.unit = unit;
		break;

	case IO_IS_PRINT:
		root->PrintStmt.unit = unit;
		break;

	case IO_IS_READ:
		root->ReadStmt.unit = unit;
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	if (io_type == IO_IS_WRITE
	|| (io_type == IO_IS_READ
		&& (root->ReadStmt.Lp != NN || root->ReadStmt.Lpeqi != NN))) {

		switch (io_type) {

		case IO_IS_WRITE:
			if (root->WriteStmt.Lpeqi != NN) {
				root->WriteStmt.Lpeqi = NN;
				root->WriteStmt.Lp = &Lp;
			}
			list = list_invert (root->WriteStmt.IoSpecList);
			break;

		case IO_IS_READ:
			if (root->ReadStmt.Lpeqi != NN) {
				root->ReadStmt.Lpeqi = NN;
				root->ReadStmt.Lp = &Lp;
			}
			list = list_invert (root->ReadStmt.IoSpecList);
			break;
		}

		spec_num = 0;

		for (LIST_LOOP (nspec, list, temp)) {
			spec_num++;
			spec = &nspec->IoSpec;

			if (spec->Name == NN) {
				switch (spec_num) {
				
				case 1:
					if (unit->unit != NN) {
						io_err_spec_used ("UNIT");
						continue;
					}

					if (spec->Expr != NN
					&& !IS_INTEGER (spec->Expr) && !IS_CHARACTER (spec->Expr)) {
						fatal (
						"expression for specifier UNIT must be of type INTEGER or CHARACTER");
						continue;
					}

					if (spec->Expr != NN)
						unit->unit = spec->Expr;
					else
						unit->unit = spec->Times;

					break;

/*!!! case 2 can also be NML, implement */

				case 2:
					if (unit->fmt != NN) {
						io_err_spec_used ("FORMAT");
						continue;
					}

					if (spec->Expr != NN) {
						if (!IS_CHARACTER (spec->Expr) && !IS_INTEGER (spec->Expr)) {
							fatal (
					"expression for specifier FMT must be of type INTEGER or CHARACTER");
							continue;
						}

						if (IS_INTEGER (spec->Expr))
							label_install_io (spec->Expr);
					}

					if (spec->Expr != NN)
						unit->fmt = spec->Expr;
					else
						unit->fmt = spec->Times;

					break;

				default:
					sprintf (serr, "keyword missing in specifier number %d", spec_num);
					fatal (serr);
				}

				continue;
			}

			name = U_strupper (spec->Name->com.str);

			if (!strncmp (name, "UNIT", 4)) {
				if (unit->unit != NN) {
					io_err_spec_used ("UNIT");
					continue;
				}

				if (spec->Expr != NN
				&& !IS_INTEGER (spec->Expr) && !IS_CHARACTER (spec->Expr)) {
					fatal (
					"expression for specifier UNIT must be of type INTEGER or CHARACTER");
					continue;
				}

				if (spec->Expr != NN)
					unit->unit = spec->Expr;
				else
					unit->unit = spec->Times;
			}
			else if (!strncmp (name, "FMT", 3)) {
					if (unit->fmt != NN) {
						io_err_spec_used ("FORMAT");
						continue;
					}

					if (spec->Expr != NN) {
						if (!IS_CHARACTER (spec->Expr) && !IS_INTEGER (spec->Expr)) {
							fatal (
					"expression for specifier FMT must be of type INTEGER or CHARACTER");
							continue;
						}

						if (IS_INTEGER (spec->Expr))
							label_install_io (spec->Expr);
					}

					if (spec->Expr != NN)
						unit->fmt = spec->Expr;
					else
						unit->fmt = spec->Times;
			}
			else if (io_spec_is ("NML", name, "IGNORE", spec, unit->nml)) {
				if (spec->Expr->com.type != NAME
				|| (entry = lookup_loc (spec->Expr->com.str)) == TN
				|| !IS_namelist (entry)) {
					fatal ("specifier NML requires a NAMELIST group name");
					continue;
				}

				unit->nml = spec->Expr;
			}
			else if (io_spec_is ("REC", name, "INTEGER", spec, unit->rec)) {
				unit->rec = spec->Expr;
			}
			else if (io_spec_is ("IOSTAT", name, "INTEGER", spec, unit->iostat)) {
				if (spec->Expr->com.type != NAME) {
					fatal ("specifier IOSTAT needs a variable"); 
					continue;
				}

				unit->iostat = spec->Expr;
			}
			else if (io_spec_is ("ERR", name, "INTEGER", spec, unit->err)) {
				if (spec->Expr->com.type != INTEGER_CONSTANT) {
					fatal ("specifier ERR needs a label");
					continue;
				}

				unit->err = spec->Expr;
			}
			else if (io_spec_is ("ADVANCE", name, "CHARACTER", spec, unit->advance)) {
				unit->advance = spec->Expr;
			}
			else {
				switch (io_type) {

				case IO_IS_WRITE:
					sprintf (serr, "invalid specifier in WRITE statement: %s", name);
					fatal (serr);
					break;

				case IO_IS_READ:
					if (io_spec_is ("END", name, "INTEGER", spec, unit->end)) {
						if (spec->Expr->com.type != INTEGER_CONSTANT) {
							fatal ("specifier END needs a label");
							continue;
						}

						unit->end = spec->Expr;
					}
					else if (io_spec_is ("SIZE", name, "INTEGER", spec, unit->size)) {
						unit->size = spec->Expr;
					}
					else if (io_spec_is ("EOR", name, "INTEGER", spec, unit->eor)) {
						if (spec->Expr->com.type != INTEGER_CONSTANT) {
							fatal ("specifier EOR needs a label");
							continue;
						}

						unit->eor = spec->Expr;
					}
					else {
						sprintf (serr, "invalid specifier in READ statement: %s", name);
						fatal (serr);
					}
					break;
				}
			}
		}
	}
	else {
		switch (io_type) {
		
		case IO_IS_READ:
			unit->fmt = root->ReadStmt.FormatSpec;
			break;

		case IO_IS_PRINT:
			unit->fmt = root->PrintStmt.FormatSpec;
			break;
		}

		if (unit->fmt != &Times) {
			if (!IS_CHARACTER (unit->fmt) && !IS_INTEGER (unit->fmt))
				fatal ("expression for format must be of type INTEGER or CHARACTER");

			if (IS_INTEGER (unit->fmt))
				label_install_io (unit->fmt);
		}
	}

	switch (io_type) {

	case IO_IS_WRITE:
		if (root->WriteStmt.OutputItemList != NN && unit->nml != NN)  /* E_9_10_42 */
			fatal (
			"NAMELIST specifier must not be present if there is an output-item-list");
		break;

	case IO_IS_READ:
		if (root->ReadStmt.InputItemList != NN && unit->nml != NN)  /* E_9_10_42 */
			fatal (
			"NAMELIST specifier must not be present if there is an input-item-list");
		break;
	}

	if (unit->nml != NN && unit->fmt != NN)  /* E_9_10_44 */
		fatal ("an io control list cannot contain both a FMT and a NML specifier");

	if (unit->rec != NN) {  /* E_9_11_11 */
		if (unit->end != NN)
			fatal ("an END specifier must not appear if a REC specifier is present");

		if (unit->nml != NN)
			fatal ("a NML specifier must not appear if a REC specifier is present");

		if (unit->fmt == &Times)
			fatal ("a '*' format must not appear if a REC specifier is present");
	}

	if (unit->eor != NN && unit->advance == NN)  /* E_9_11_17 */
		fatal ("an ADVANCE specifier must apper if a EOR specifier is present");
	
	return root;
}

checkx_FormatStmt (sta)
Node *sta;
{
	if (flag_cubix)
		cubix_checkx_FormatStmt (sta);
}

/********************************* W R I T E ********************************/

Node *trans_WriteStmt (root)
Node *root;
{
	return trans_IoStmt (root, IO_IS_WRITE);
}

Node *trans_ReadStmt (root)
Node *root;
{
	return trans_IoStmt (root, IO_IS_READ);
}

Node *trans_PrintStmt (root)
Node *root;
{
	return trans_IoStmt (root, IO_IS_PRINT);
}

Node *trans_IoStmt (root, io_type)
Node *root;
int io_type;
{
	Node *imp_val_list;

	switch (io_type) {

	case IO_IS_WRITE:
		imp_val_list = root->WriteStmt.OutputItemList;
		break;

	case IO_IS_PRINT:
		imp_val_list = root->PrintStmt.OutputItemList;
		break;

	case IO_IS_READ:
		imp_val_list = root->ReadStmt.InputItemList;
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	io_expand_imp_val_list (imp_val_list, io_type);

	return root;
}

io_expand_imp_val_list (imp_val_list, io_type)
Node *imp_val_list;
int io_type;
{
	Node *imp_val, *temp, *temp1, *arr;
	double value;

	for (LIST_LOOP (imp_val, imp_val_list, temp)) {
		int was_dum_expr = 0, typshapeeq;
		T_Entry *temp_entry = TN;

		switch (imp_val->com.type) {

		default:
		case DUM_EXPR:
		case EXPR:
			if (imp_val->com.typ->shape != SN
			&&  imp_val->com.typ->shape->rank) {
				if (io_type == IO_IS_READ)
					temp->lst.value = temp_replace_inout (imp_val);
				else
					temp->lst.value = temp_replace (imp_val);

				temp_entry = lookup_loc (temp->lst.value->com.str);
			}

			was_dum_expr = 1;

			break;

		case IMPLIED_DO:
			io_expand_imp_val_list (imp_val->ImpliedDo.ImpliedValueList, io_type);
			break;

		case ARRAY_SECTION:
		{
			Node *sub;
			T_Entry *entry;

			entry = lookup_loc (imp_val->com.str);
			if (entry->typ->shape == SN || entry->typ->shape->rank == 0)
				break;  /* catches character*10 cha; read (10, *) cha (1:4) */

			if (IS_alloc_temp (entry)
			||
			(	IS_automatic (entry)
				&& imp_val->com.typ->shape && imp_val->com.typ->shape->rank)) {
					if (io_type == IO_IS_READ)
						temp->lst.value = temp_replace_inout (imp_val);
					else
						temp->lst.value = temp_replace (imp_val);

					temp_entry = lookup_loc (temp->lst.value->com.str);

					was_dum_expr = 1;
					break;
			}
			
			for (LIST_LOOP (sub, imp_val->ArraySection.SubscriptList, temp1)) {
				if (sub->Subscript.Expr->com.typ->shape != SN
				&&  sub->Subscript.Expr->com.typ->shape->rank) {
					if (io_type == IO_IS_READ)
						temp->lst.value = temp_replace_inout (imp_val);
					else
						temp->lst.value = temp_replace (imp_val);

					temp_entry = lookup_loc (temp->lst.value->com.str);

					was_dum_expr = 1;
					break;
				}
			}
		}
			break;

		case STRIDE:  /*!!! should give syntax error??? */
			break;
		}

		arr = temp->lst.value;

		if (arr->com.type == ARRAY_SECTION)
			typshapeeq = typ_shape_eq (arr->com.typ->shape
			                        , (lookup_loc (arr->com.str))->typ->shape);

		if (arr->com.type == ARRAY_SECTION
		&&  arr->com.typ->shape != SN  /* catches arr (1, 2) */
		&& (was_dum_expr && temp_entry != TN && IS_alloc_temp (temp_entry) || !typshapeeq)) {
			Node *imp_do, *imp_do_cont, *imp_val_list, *sub, *temp1;
			Typ *typ;
			Shape *shape = arr->com.typ->shape;
			Node *var, *temps [7], *dummy;
			int i, num, numsubs;

			/* replaces subscripts with extent by loop index */

			num = 0;
			for (LIST_LOOP (sub, arr->ArraySection.SubscriptList, temp1)) {
				if (sub->Subscript.Colon != NN) {
					typ = typ_malloc ();
					typ->etype = Etype_INTEGER_0;
					temps [num] = temp_make (typ, 1, &dummy);

					num++;
				}
			}

			numsubs = num;
			num = 0;
			for (LIST_LOOP (sub, arr->ArraySection.SubscriptList, temp1)) {
				if (sub->Subscript.Colon != NN) {
					temp1->lst.value = make_Subscript (temps [numsubs - num - 1], NN, NN);
					num++;
				}
			}

			/* makes nested implicit loops */

			imp_val_list = make_List (&(temp->lst.value), 1, IMPLIED_VALUE_LIST);

			for (i = 0; i < num; i++) {
				imp_do_cont = make_ImpliedDoControl (temps [i]
				, shape->sizes0 [i], shape->sizes1 [i], shape->sizes2 [i]);

				imp_val_list = make_ImpliedDo (imp_val_list, imp_do_cont);
			}

			temp->lst.value = imp_val_list;
		}
	}
}

/********************************* W R I T E ********************************/

static io_is_not_to_string (node, io_type)
Node *node;
int io_type;
{
	Unit *unit;

	switch (io_type) {

	case IO_IS_WRITE:
		unit = node->WriteStmt.unit;
		break;

	case IO_IS_READ:
		unit = node->WriteStmt.unit;
		break;

	case IO_IS_PRINT:
		unit = node->WriteStmt.unit;
		break;

#ifdef CHECK
	default:
		SYSERR;
#endif
	}

	if (unit->unit == NN)
		return 1;

	if (unit->unit->com.typ != (Typ*) 0
	&& unit->unit->com.typ->etype->type == TYP_CHARACTER)
		return 0;
	else
		return 1;
}

write_IOkeyword (node)
Node *node;
{
	io_in = 1;
	ostr (node_Symbol (node->com.type));
}

write_WriteStmt (node)
Node *node;
{
	if (flag_cubix && io_is_not_to_string (node, IO_IS_WRITE))
		cubix_write_IoStmt (node, IO_IS_WRITE);
	else
		ochl (node);
}

write_ReadStmt (node)
Node *node;
{
	if (flag_cubix)
		cubix_write_IoStmt (node, IO_IS_READ);
	else
		ochl (node);
}

write_PrintStmt (node)
Node *node;
{
	if (flag_cubix)
		cubix_write_IoStmt (node, IO_IS_PRINT);
	else
		ochl (node);
}
